import json

import spotipy

from spotcrates.playlists import Playlists


def get_spotify_handle():
    cache_handler = spotipy.cache_handler.CacheFileHandler()
    auth_manager = spotipy.oauth2.SpotifyOAuth(cache_handler=cache_handler, scope=["playlist-modify-private",
                                                                                   "playlist-read-private"])
    return spotipy.Spotify(auth_manager=auth_manager)


def print_all_playlists():
    sp = get_spotify_handle()

    playlists = sp.current_user_playlists()
    while playlists:
        for key, value in playlists.items():
            print(f"Playlists entry: {key} -> {value}")
        for i, playlist in enumerate(playlists['items']):
            for key, value in playlist.items():
                print(f"Playlist entry: {key} -> {value}")
            print("%4d %s %s" % (i + 1 + playlists['offset'], playlist['uri'], playlist['name']))
        if playlists['next']:
            playlists = sp.next(playlists)
        else:
            playlists = None


def run_append_daily():
    sp = get_spotify_handle()

    playlists = Playlists(sp)
    playlists.append_daily_mix()


def run_list_all():
    sp = get_spotify_handle()

    playlists = Playlists(sp)
    print(json.dumps(playlists.get_all_playlists()))


def print_tracks():
    sp = get_spotify_handle()

    playlist_items = sp.playlist_items("20HoheMV2htBPBX07D8pHm")
    print_playlist_page(playlist_items)

    # next_playlist_page = sp.next(playlist_items)
    # print_playlist_page(next_playlist_page)

    track_ids = {playlist_item['track']['id'] for playlist_item in playlist_items['items']}

    next_playlist_page = sp.next(playlist_items)
    while next_playlist_page:
        track_ids.update({playlist_item['track']['id'] for playlist_item in next_playlist_page['items']})
        next_playlist_page = sp.next(next_playlist_page)

    print(f"Track IDs type: {type(track_ids)}")
    print(f"Track IDs size: {len(track_ids)}")
    print(f"Track IDs: {track_ids}")
    # for i, playlist_item in enumerate(playlist_items['items']):
    #     print("=====================%4d %s %s=====================" % (i + 1, playlist_item['added_at'],
    #                                                                    playlist_item['track']['id']))

    # track = playlist_item['track']
    # for key, value in track.items():
    #     print(f"Track entry: {key} -> {value}")
    #
    #
    # for key, value in playlist_item.items():
    #     print(f"Playlist item entry: {key} -> {value}")


def print_playlist_page(playlist_items):
    for key, value in playlist_items.items():
        # print(f"Playlists key: {key}")
        # print(f"Playlists key: {type(key)}")
        # print(f"{key} == 'items'? {key == 'items'}")
        if key == 'items':
            print("Skipping items")
            continue
        print(f"Playlists Item: {key} -> {value}")


if __name__ == '__main__':
    # print_all_playlists()
    # print_tracks()
    run_append_daily()
    #run_list_all()



"""
Prerequisites

    pip3 install spotipy Flask Flask-Session

    // from your [app settings](https://developer.spotify.com/dashboard/applications)
    export SPOTIPY_CLIENT_ID=client_id_here
    export SPOTIPY_CLIENT_SECRET=client_secret_here
    export SPOTIPY_REDIRECT_URI='http://127.0.0.1:8080' // must contain a port
    // SPOTIPY_REDIRECT_URI must be added to your [app settings](https://developer.spotify.com/dashboard/applications)
    OPTIONAL
    // in development environment for debug output
    export FLASK_ENV=development
    // so that you can invoke the app outside of the file's directory include
    export FLASK_APP=/path/to/spotipy/examples/app.py
 
    // on Windows, use `SET` instead of `export`

Run app.py

    python3 app.py OR python3 -m flask run
    NOTE: If receiving "port already in use" error, try other ports: 5000, 8090, 8888, etc...
        (will need to be updated in your Spotify app and SPOTIPY_REDIRECT_URI variable)
"""

import os
from flask import Flask, session, request, redirect, render_template
from flask_session import Session
import spotipy

app = Flask(__name__)
app.config['SECRET_KEY'] = os.urandom(64)
app.config['SESSION_TYPE'] = 'filesystem'
app.config['SESSION_FILE_DIR'] = './.flask_session/'
Session(app)


@app.route('/')
def index():
    cache_handler = spotipy.cache_handler.FlaskSessionCacheHandler(session)
    auth_manager = spotipy.oauth2.SpotifyOAuth(scope='user-read-currently-playing playlist-modify-private',
                                               cache_handler=cache_handler,
                                               show_dialog=True)

    if request.args.get("code"):
        # Step 2. Being redirected from Spotify auth page
        auth_manager.get_access_token(request.args.get("code"))
        return redirect('/')

    if not auth_manager.validate_token(cache_handler.get_cached_token()):
        # Step 1. Display sign in link when no token
        context = {
            'auth_url': auth_manager.get_authorize_url()
        }

        return render_template('index.html', **context)

    # Step 3. Signed in, display data
    spotify = spotipy.Spotify(auth_manager=auth_manager)
    context = {
        'spotify': spotify
    }
    return render_template('auth_index.html', **context)


@app.route('/sign_out')
def sign_out():
    session.pop("token_info", None)
    return redirect('/')


@app.route('/playlists')
def playlists():
    cache_handler = spotipy.cache_handler.FlaskSessionCacheHandler(session)
    auth_manager = spotipy.oauth2.SpotifyOAuth(cache_handler=cache_handler)
    if not auth_manager.validate_token(cache_handler.get_cached_token()):
        return redirect('/')

    spotify = spotipy.Spotify(auth_manager=auth_manager)
    return spotify.current_user_playlists()


@app.route('/currently_playing')
def currently_playing():
    cache_handler = spotipy.cache_handler.FlaskSessionCacheHandler(session)
    auth_manager = spotipy.oauth2.SpotifyOAuth(cache_handler=cache_handler)
    if not auth_manager.validate_token(cache_handler.get_cached_token()):
        return redirect('/')
    spotify = spotipy.Spotify(auth_manager=auth_manager)
    track = spotify.current_user_playing_track()
    if not track is None:
        return track
    return "No track currently playing."


@app.route('/current_user')
def current_user():
    cache_handler = spotipy.cache_handler.FlaskSessionCacheHandler(session)
    auth_manager = spotipy.oauth2.SpotifyOAuth(cache_handler=cache_handler)
    if not auth_manager.validate_token(cache_handler.get_cached_token()):
        return redirect('/')
    spotify = spotipy.Spotify(auth_manager=auth_manager)
    return spotify.current_user()


'''
Following lines allow application to be run more conveniently with
`python app.py` (Make sure you're using python3)
(Also includes directive to leverage pythons threading capacity.)
'''
if __name__ == '__main__':
    app.run(threaded=True, port=int(os.environ.get("PORT",
                                                   os.environ.get("SPOTIPY_REDIRECT_URI", 8080).split(":")[-1])))



#!/usr/bin/env python

"""
CLI runner for Spotify automation.
"""

import sys
import argparse
import logging

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

__author__ = 'cmayes'

from pathlib import Path

import spotipy

from spotcrates.playlists import Playlists
from appdirs import user_config_dir

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)

DEFAULT_CONFIG_DIR = user_config_dir('spotcrates')
DEFAULT_CONFIG_FILE = Path(DEFAULT_CONFIG_DIR, "spotcrates_config.toml")

COMMANDS = ['daily']

COMMAND_DESCRIPTION = f"""
{'COMMAND NAME':<16} DESCRIPTION
{'daily':<16} Add "Daily Mix" entries to the end of the target playlist, filtering for excluded entries.
"""


def warning(*objs):
    """Writes a message to stderr."""
    print("WARNING: ", *objs, file=sys.stderr)


def get_config(config_file):
    if config_file.exists():
        with open(config_file, mode="rb") as fp:
            return tomllib.load(fp)
    else:
        logging.debug(f"Config file '{config_file}' does not exist")
        return {}


def print_commands():
    """Prints available commands."""
    print(COMMAND_DESCRIPTION)


def get_spotify_handle(config):
    cache_handler = spotipy.cache_handler.CacheFileHandler()
    spotify_cfg = config.get("spotify")
    if spotify_cfg:
        auth_manager = spotipy.oauth2.SpotifyOAuth(client_id=spotify_cfg.get("client_id"),
                                                   client_secret=spotify_cfg.get("client_secret"),
                                                   redirect_uri=spotify_cfg.get("redirect_uri"),
                                                   cache_handler=cache_handler,
                                                   scope=["playlist-modify-private", "playlist-read-private"])
    else:
        auth_manager = spotipy.oauth2.SpotifyOAuth(cache_handler=cache_handler,
                                                   scope=["playlist-modify-private", "playlist-read-private"])
    return spotipy.Spotify(auth_manager=auth_manager)


def append_daily_mix(config):
    sp = get_spotify_handle(config)

    playlists = Playlists(sp)
    playlists.append_daily_mix()


def parse_cmdline(argv):
    """
    Returns the parsed argument list and return code.
    `argv` is a list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    # initialize the parser object:
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config_file", help="The location of the config file",
                        default=DEFAULT_CONFIG_FILE, type=Path)
    parser.add_argument("command", metavar='COMMAND',
                        help=f"The command to run (one of {','.join(COMMANDS)})")
    args = None
    try:
        args = parser.parse_args(argv)
    except IOError as e:
        warning("Problems reading file:", e)
        parser.print_help()
        return args, 2

    return args, 0


def main(argv=None):
    args, ret = parse_cmdline(argv)
    if ret != 0:
        return ret
    config = get_config(args.config_file)
    command = args.command.lower()

    if command == 'daily':
        append_daily_mix(config)
    elif command == 'commands':
        print_commands()
    else:
        print(f"Invalid command '{args.command}'.  Valid commands: {','.join(COMMANDS)}")
        return 1

    return 0  # success


if __name__ == '__main__':
    status = main()
    sys.exit(status)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spotcrates']

package_data = \
{'': ['*'], 'spotcrates': ['templates/*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'flask-session>=0.4.0,<0.5.0',
 'flask>=2.2.2,<3.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'spotipy>=2.22.0,<3.0.0']

entry_points = \
{'console_scripts': ['spotcrates = spotcrates.cli:main']}

setup_kwargs = {
    'name': 'spotcrates',
    'version': '0.1.2',
    'description': '',
    'long_description': '# spotcrates\n\nTools for finding and sharing Spotify tracks.',
    'author': 'Chris Mayes',
    'author_email': 'cmayes@cmay.es',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import logging
from itertools import islice
from typing import List, Set

from spotipy import Spotify

DAILY_MIX_PREFIX = 'Daily Mix'
DAILY_MIX_TARGET = 'Now'
DAILY_MIX_EXCLUDES = 'Overplayed'


class PlaylistException(Exception):
    pass


def batched(iterable, n):
    "Batch data into lists of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    it = iter(iterable)
    while True:
        batch = list(islice(it, n))
        if not batch:
            return
        yield batch


class Playlists:
    def __init__(self, spotify: Spotify):
        self.spotify = spotify
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)

    def get_all_playlists(self) -> dict:
        return self.spotify.current_user_playlists()['items']

    def append_daily_mix(self, target_list_name=DAILY_MIX_TARGET, daily_mix_prefix=DAILY_MIX_PREFIX,
                         exclude_list_name=DAILY_MIX_EXCLUDES):
        dailies = []
        target_list = None
        exclude_list = None
        for playlist in self.get_all_playlists():
            list_name = playlist['name']
            if list_name:
                if list_name.startswith(daily_mix_prefix):
                    dailies.append(playlist)
                elif list_name == target_list_name:
                    target_list = playlist
                elif list_name == exclude_list_name:
                    exclude_list = playlist

        # TODO: Optionally create if it doesn't exist
        if not target_list:
            raise PlaylistException(f"Could not find target list '{target_list_name}")

        # user_playlist_add_tracks(user, playlist_id, tracks, position=None)
        if not dailies:
            self.logger.warning(f"No daily mixes found with the prefix '{daily_mix_prefix}'")

        exclude_ids = self._get_playlist_track_ids(target_list["id"])

        if exclude_list:
            exclude_ids.update(self._get_playlist_track_ids(exclude_list["id"]))

        add_tracks = []
        orig_daily_count = 0
        for daily in dailies:
            daily_items = self._get_playlist_tracks(daily["id"])
            orig_daily_count += len(daily_items)
            add_tracks.extend(
                [daily_item for daily_item in daily_items if daily_item['track']['id'] not in exclude_ids])

        self.logger.info(f"{len(add_tracks)} to add from an original count of {orig_daily_count}")
        if add_tracks:
            self._add_tracks_to_playlist(target_list, add_tracks)
        else:
            self.logger.warning("No daily songs to add")

    def _get_playlist_track_ids(self, *args: str) -> Set[str]:
        track_ids = set()
        for playlist_id in args:
            # TODO: See about paring down to just the ID via "fields" param
            playlist_items = self.spotify.playlist_items(playlist_id)

            track_ids.update({playlist_item['track']['id'] for playlist_item in playlist_items['items']})

            next_playlist_page = self.spotify.next(playlist_items)
            while next_playlist_page:
                track_ids.update({playlist_item['track']['id'] for playlist_item in next_playlist_page['items']})
                next_playlist_page = self.spotify.next(next_playlist_page)

        return track_ids

    def _get_playlist_tracks(self, *args: str) -> List[dict]:
        tracks = []
        for playlist_id in args:
            # TODO: See about paring down to just the ID via "fields" param
            playlist_items = self.spotify.playlist_items(playlist_id)
            tracks.extend(playlist_items['items'])

            next_playlist_page = self.spotify.next(playlist_items)
            while next_playlist_page:
                tracks.extend(next_playlist_page['items'])
                next_playlist_page = self.spotify.next(next_playlist_page)

        return tracks

    def _add_tracks_to_playlist(self, target_list, add_tracks):
        track_ids = {add_song['track']['id'] for add_song in add_tracks}

        for id_batch in batched(track_ids, 100):
            self.logger.debug(f"Batch size: {len(id_batch)}")
            self.spotify.playlist_add_items(target_list["id"], id_batch)
